/* ########################################################################

				 Type.c

   File: Type.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/Type.c
   Description: 
   Created: Tue Feb 21 12:46:10 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 12:46:10 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */



/******************************** Les types **********************************/

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include <stdio.h>

#include "Type_.h"
#include "type.h"
#include "hash.h"
#include "Instruction_.h"
#include "error.h"

/* Les types builtins */

Type * Type_Int;
Type * Type_Char;
Type * Type_Void;

Type * Type_String;
Type * Type_PInt;
Type * Type_PVoid;

/* to memorize all the types */

typedef struct TypeList {
  Type * type;
  struct TypeList * next;
} TypeList;

/**/

Type * Type__Type(subinfo, tot, size_of)
    void * subinfo;
    type_of_type tot;
    int size_of;
{
  /* all the types */
  static TypeList * all[T_Function + 1];
  
  Type * this = (Type *) Malloc(sizeof(Type));
  TypeList * l;

  this->_name_or_pointed = subinfo;
  this->_type = tot;
  this->_sizeof = size_of;
  this->_dontfree = 0;

  /* is the type already exist ? */
  for (l = all[tot]; l; l = l->next) {
    if (Type__Eq(l->type, this)) {
      /* find ! */
      Type__Delete(this);
      return l->type;
    }
  }
  
  /* new type */
  this->_dontfree = 1;
  
  l = (TypeList *) Malloc(sizeof(TypeList));
  l->type = this;
  l->next = all[tot];
  all[tot] = l;
    
  return this;
}

/* delete a type */

void Type__Delete(this)
    Type * this;
{
  Type * cont;
  
  for (;;) {
    if (this->_dontfree)
      return;
    
    if (this->_type == T_Function){
      TypeFunction * tf = Type__Function(this);
      Type ** argtypes = &TypeFunction__ArgType(tf, 0);
      int narg = TypeFunction__NbreArg(tf);
      
      cont = TypeFunction__ValType(tf);
      
      while (narg--)
	Type__Delete(argtypes[narg]);

      /* do NOT free argtypes ! */
      
      free(tf);
    }
    else
      /* T_Pointer or T_Array (cannot be T_Builtin : _dontfree) */
      cont = Type__Pointed_Type(this);
    
    free(this);
    this = cont;
  }
}

int Type__Equal(t1, t2)
     Type * t1;
     Type * t2;
{
  int firstdim = 1;
  
  for (;;) {
    if (t1 == t2)
      return 1;

    if ((t1->_type == T_Builtin) || (t2->_type == T_Builtin))
      return 0;

    if (t1->_type == T_Function) {
      if (t2->_type == T_Function) {
	TypeFunction * tf1 = Type__Function(t1);
	TypeFunction * tf2 = Type__Function(t2);

	if ((tf1->_nbre_arg == tf2->_nbre_arg) &&
	    Type__Equal(tf1->_valtype, tf2->_valtype)) {
	  Type ** at1 = tf1->_argtypes;
	  Type ** at2 = tf2->_argtypes;
	  int index = tf1->_nbre_arg;
	
	  while (index--)
	    if (! Type__Equal(at1[index], at2[index]))
	      return 0;
	  return 1;
	}
      }
      return 0;
    }
    else if (t2->_type == T_Function)
      return 0;
    else if (firstdim)
      firstdim = 0;
    else if ((t1->_type != t2->_type) ||
	     (t1->_sizeof != t2->_sizeof))
      return 0;
    
    t1 = (Type *) t1->_name_or_pointed;
    t2 = (Type *) t2->_name_or_pointed;
  }
}

int Type__Eq(t1, t2)
     Type * t1;
     Type * t2;
{
  for (;;) {
    if (t1 == t2)
      return 1;

    if ((t1->_type != t2->_type) ||
	(t1->_sizeof != t2->_sizeof))
      return 0;
    
    if (t1->_type == T_Builtin)
      return 0;

    if (t1->_type == T_Function) {
      TypeFunction * tf1 = Type__Function(t1);
      TypeFunction * tf2 = Type__Function(t2);
      
      if ((tf1->_nbre_arg == tf2->_nbre_arg) &&
	  Type__Eq(tf1->_valtype, tf2->_valtype)) {
	Type ** at1 = tf1->_argtypes;
	Type ** at2 = tf2->_argtypes;
	int index = tf1->_nbre_arg;
	
	while (index--)
	  if (! Type__Eq(at1[index], at2[index]))
	    return 0;
	return 1;
      }
      return 0;
    }
    
    t1 = (Type *) t1->_name_or_pointed;
    t2 = (Type *) t2->_name_or_pointed;
  }
}


/* Ajoute le nom du type a la fin de str */

void Type__Print(t, str)
    Type * t;
    char * str;
{
  if (Type__IsBuiltin(t))
    strcat(str, Type__Name(t));
  else if (Type__IsaFunction(t)) {
    TypeFunction * tf = Type__Function(t);
    
    Type__Print(TypeFunction__ValType(tf), str);
    strcat(str, "(*)(");
    if (TypeFunction__NbreArg(tf) &&
	tf->_argtypes) {	/* operators does not have argtypes */
      int i;
      
      Type__Print(TypeFunction__ArgType(tf, 0), str);
      for (i = 1; i < TypeFunction__NbreArg(tf); i += 1) {
	strcat(str, ", ");
	Type__Print(TypeFunction__ArgType(tf, i), str);
      }
    }
    strcat(str, ")");
  }
  else {
    Type__Print(Type__Pointed_Type(t), str);
    strcat(str, "*");
  }
}


/* Compatibilite de deux types

   retourne 0 si t1 ne peut etre transcrit en t2
   Type__Convertible si les types sont interchangeables
   une fonction de conversion sinon */

HashTable * convert_hash_table;

ConvFct Type__Convertible(t1, t2)
     Type * t1;
     Type * t2;
{
  if (Type__Equal(t1, t2))
    return (ConvFct) Type__Convertible;

  {
    /* type * convertible en void * */
    
    Type * st1 = t1;
    Type * st2 = t2;

    while (((st1->_type ==  T_Array) || (st1->_type == T_Pointer)) &&
	   (st2->_type == T_Pointer)) {
      st1 = Type__Pointed_Type(st1);
      st2 = Type__Pointed_Type(st2);
      if (st2 == Type_Void)
	return AlreadyConvertible;
    }
    
    if ((st1->_type == T_Function) &&
	(st2->_type == T_Pointer) &&
	(Type__Pointed_Type(st2) == Type_Void))
      return AlreadyConvertible;
  }
  
  {
    /* Regarde s'il y a une conversion possible */
    
    char encoded[24];
    
    sprintf(encoded, "%s2", encode_type(t1));
    strcat(encoded, encode_type(t2));

    return (ConvFct) HashTable__Search(convert_hash_table, (Object) encoded);
  }
}


/* La meme chose mais avec la forme ce qui permet de tester le cas
   0, valeur pouvant etre affectee a n'importe quel type d'objet */

ConvFct ExprType__Convertible(i, t)
     Instruction * i;
     Type * t;
{
  Type * ti = GetExprType(i);
  
  if ((ti == Type_Int) && IsaConst(i) && (! Eval(i)))
    return AlreadyConvertible;
  
  return Type__Convertible(ti, t);
}


/* Les fonctions de conversion */

char int2char(i)
     int i;
{
  return (char) i;
}

int char2int(c)
     int c;
{
  /* Cette fonction est la parce que l'on ne
     sait pas si un char est ou non signe */
  return (int) ((char) c);
}


/* Sort en erreur pour incompatibilite de deux types */

void error_incompatible_types(t1, t2, msg)
    Type * t1;
    Type * t2;
    char * msg;
{
  strcpy(err_msg, msg);
  strcat(err_msg, ", ");
  Type__Print(t1, err_msg);
  strcat(err_msg, " <> ");
  Type__Print(t2, err_msg);
  
  Error(err_msg, 1);
}

/* Conv */

static inherit_instruction_function_vector
  conv_inherit_instruction_function_vector;

Instruction * Conv__Conv(expr, conv)
    Instruction * expr;
    ConvFct conv;
{
  Conv * this;
  
  if (conv == AlreadyConvertible)
    /* conversion useless */
    return expr;
 
  this = (Conv *) Malloc(sizeof(Conv));

  Instruction_Constructor
    (conv_inherit_instruction_function_vector, this,
     (conv == (ConvFct) int2char) ? Type_Char : Type_Int);
  
  this->_conv = conv;
  this->_expr = expr;
  
  if (conv == 0) conv(0);
  
  return (Instruction *) this;
}

static void Conv__Delete(this)
     Conv * this;
{
  Delete(this->_expr);
  free(this);
}

static Object Conv__Eval(this)
     Conv * this;
{
  return this->_conv(Eval(this->_expr));
}

/* A appeler dans main */

void Init_type()
{
  Type_Int =  Type__Type("int", T_Builtin, sizeof(int));
  Type_Char = Type__Type("char", T_Builtin, sizeof(char));
  Type_Void = Type__Type("void", T_Builtin, sizeof(void *));

  Type_String = Type__Type(Type_Char, T_Pointer, sizeof(char *));
  Type_PInt = Type__Type(Type_Int, T_Pointer, sizeof(int *));
  Type_PVoid = Type__Type(Type_Void, T_Pointer, sizeof(void *));

  /**/
  
  convert_hash_table = HashTable__HashTable(24, 1);
  
  if (sizeof(char) != sizeof(int)) {
    HashTable__Add(convert_hash_table, (Object) "int2char", (Object) int2char);
    HashTable__Add(convert_hash_table, (Object) "char2int", (Object) char2int);
  }
  
  conv_inherit_instruction_function_vector._Eval = (EvalFct) Conv__Eval;
  conv_inherit_instruction_function_vector._Delete = (DeleteFct) Conv__Delete;
}
