/* ########################################################################

				scroll.c

   File: scroll.c
   Path: /home/fournigault/c/X11/xcoral-2.31/scroll.c
   Description: 
   Created: Fri Jan 27 11:28:01 MET 1995
   Author: Lionel Fournigault
   Modified: Fri Jan 27 11:28:02 MET 1995
   Last maintained by: Lionel Fournigault

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Lionel Fournigault

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */
/* #define DEBUG */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include <X11/Xutil.h>

#if STDC_HEADERS
#  include <stdlib.h>
#else
#  if HAVE_MALLOC_H
#    include <malloc.h>
#  endif
#endif

#include "scroll.h"
#include "shadow.h"
#include "parse.h"
#include "buttons_bits.h"

static char gray [] = { 0x1, 0x02 };

static Pixmap	pix;
static Cursor   cursor;
static unsigned long gray_color1;
static unsigned long gray_color2;
static unsigned long gray_color3;
static unsigned long gray_color4;

#define GRAY1 "gray75"
#define GRAY2 "gray35"
#define GRAY3 "gray47"
#define GRAY4 "gray86"

static Pixmap sb_up_pix;
static Pixmap sb_down_pix;
static Pixmap sb_up_pix_run;
static Pixmap sb_down_pix_run;
static XImage * sb_up_im;
static XImage * sb_down_im;
static XImage * sb_up_im_run;
static XImage * sb_down_im_run;

static GC sb_mess_gc;

FCT (static int, myrint, (double x) );
extern XImage *LoadImage256 (char **t);
/*
**	Function name : InitScroll
**
**	Description : Pour l'instant, on cree le cursor associe a
**		la barre de scroll, et les pixmap pour les boutons up down
**
**	Input :  Le display.
**	Ouput :
*/
void InitScroll ( display )
    Display	*display;
{
  /* cursor */
  {
    int result;
    pix = XCreatePixmapFromBitmapData ( display, 
				       DefaultRootWindow (display), gray, 2, 2, 
				       BlackPixel ( display, DefaultScreen (display)),
				       WhitePixel ( display, DefaultScreen (display)),
				       DefaultDepth (display, DefaultScreen ( display )));
    
    cursor = XCreateFontCursor ( display, XC_sb_left_arrow );
    
    gray_color1 = PixelValue ( display, GRAY1, &result );
    if ( result == False )
      gray_color1 = BlackPixel ( display, DefaultScreen (display));
    gray_color2 = PixelValue ( display, GRAY2, &result );
    if ( result == False )
      gray_color2 = WhitePixel ( display, DefaultScreen (display));
    gray_color3 = PixelValue ( display, /*"dimgray"*/ GRAY3, &result );
    if ( result == False )
      gray_color3 = BlackPixel ( display, DefaultScreen (display));
    gray_color4 = PixelValue ( display, /*"gainsboro"*/ GRAY4, &result );
    if ( result == False )
      gray_color4 = WhitePixel ( display, DefaultScreen (display));
  }
  
  /* buttons */
  {
    XGCValues 	gcv;
    Window	root;
    int 	depth;
    
    root = DefaultRootWindow ( display );
    depth = DefaultDepth ( display, DefaultScreen ( display ));
    sb_mess_gc = XCreateGC ( display, root, 0, &gcv );
    
    if ((depth == 16)||(depth == 24)) {
      sb_up_im = (XImage *)LoadImage256 ((char **) xpm_sb_up_bits);
      sb_up_pix = XCreatePixmap (display, DefaultRootWindow (display),
				 SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT, depth);
      XPutImage (display, sb_up_pix, sb_mess_gc, sb_up_im, 0, 0, 0, 0,
		 SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT);
      
      sb_down_im = (XImage *)LoadImage256 ((char **) xpm_sb_down_bits);
      sb_down_pix = XCreatePixmap (display, DefaultRootWindow (display),
				   SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT, depth);
      XPutImage (display, sb_down_pix, sb_mess_gc, sb_down_im, 0, 0, 0, 0,
		 SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT);

      sb_down_im_run = (XImage *)LoadImage256 ((char **) xpm_sb_down_bits_run);
      sb_down_pix_run = XCreatePixmap (display, DefaultRootWindow (display),
				   SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT, depth);
      XPutImage (display, sb_down_pix_run, sb_mess_gc, sb_down_im_run, 0, 0, 0, 0,
		 SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT);

      sb_up_im_run = (XImage *)LoadImage256 ((char **) xpm_sb_up_bits_run);
      sb_up_pix_run = XCreatePixmap (display, DefaultRootWindow (display),
				 SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT, depth);
      XPutImage (display, sb_up_pix_run, sb_mess_gc, sb_up_im_run, 0, 0, 0, 0,
		 SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT);
    }
    else {
      sb_up_pix = 
	XCreatePixmapFromBitmapData ( display, root, (char *) sb_up_bits,
				     SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT,
				     gray_color1, gray_color2, depth );
      sb_down_pix =
	XCreatePixmapFromBitmapData ( display, root, (char *) sb_down_bits,
				     SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT,
				     gray_color1, gray_color2, depth );
      sb_up_pix_run = 
	XCreatePixmapFromBitmapData ( display, root, (char *) sb_up_bits_run,
				     SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT,
				     gray_color1, gray_color2, depth );
      sb_down_pix_run =
	XCreatePixmapFromBitmapData ( display, root, (char *) sb_down_bits_run,
				     SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT,
				     gray_color1, gray_color2, depth );
    }
  }
}


/*
**	Function name : MakeScroll
**
**	Description :  Construction d'une fenetre de scroll avec la
**		scrollbar.
**
**	Input :  Le display, la fenetre parent, position dans la fenetre
**		parent.
**	Ouput : Un pointeur sur SWin
*/
SWin *MakeScroll ( display, parent, x, y )
    Display	*display;
    Window	parent;
    int x, y;
{
    SWin *s;
    
    s = (SWin *) malloc ( (unsigned) sizeof ( SWin )); 
    
    s -> frame = XCreateSimpleWindow ( display, parent,
				      x, y, SCROLL_WIDTH - 2, DEFAULT_SIZE, 0,
				      BlackPixel ( display, DefaultScreen (display)),
				      WhitePixel ( display, DefaultScreen (display)));
    
    XSelectInput ( display, s -> frame, ButtonPressMask | ButtonMotionMask
		  | ButtonReleaseMask | ExposureMask | PointerMotionHintMask );
    
    s -> fx = x; s -> fy = y;
    
    s -> scroll = XCreateSimpleWindow ( display, s -> frame,
				       2, 2 + SCROLL_BUTTONS_HEIGHT,
				       SCROLL_WIDTH - 4 , DEFAULT_SIZE, 0,
				       BlackPixel ( display, DefaultScreen (display)),
				       WhitePixel ( display, DefaultScreen (display)));
    
    s -> sx = 2;
    s -> sy = 2 + SCROLL_BUTTONS_HEIGHT;
    s -> line = s -> linepage = 0;
    s -> line_to_scroll = 0;
    s -> rest = 0;
    s -> last_dir = 0;
    
    XSelectInput ( display, s -> scroll, ExposureMask );
    s -> width = SCROLL_WIDTH;
    
    if ( DefaultDepth ( display, DefaultScreen (display)) == 1 ) {
	/* Black and White */
	XSetWindowBackgroundPixmap ( display, s -> scroll, pix );
	s -> bg = BlackPixel ( display, DefaultScreen (display));
	s -> fg = WhitePixel ( display, DefaultScreen (display));
    }
    else {
      /* Color */
      XSetWindowBackground ( display, s -> scroll, gray_color1 );
      XSetWindowBackground ( display, s -> frame, gray_color2 );
      s -> bg = gray_color4;
      s -> fg = gray_color3;
    }
    
    s->up = XCreateSimpleWindow ( display, s -> frame,
				 2, 2, SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT, 0,
				 BlackPixel ( display, DefaultScreen (display)),
				 WhitePixel ( display, DefaultScreen (display)) );
    XSetWindowBackgroundPixmap ( display, s -> up, sb_up_pix );
    XSelectInput ( display, s->up, ExposureMask
		  | ButtonPressMask | ButtonReleaseMask );
    
    s->down = XCreateSimpleWindow ( display, s -> frame,
				   2, 2, SCROLL_BUTTONS_WIDTH, SCROLL_BUTTONS_HEIGHT, 0,
				   BlackPixel ( display, DefaultScreen (display)),
				   WhitePixel ( display, DefaultScreen (display)) );
    XSetWindowBackgroundPixmap ( display, s -> down, sb_down_pix );
    XSelectInput ( display, s->down, ExposureMask
		  | ButtonPressMask | ButtonReleaseMask );
    
    return ( s );
}


/*
**	Function name : DeleteScroll
**
**	Description : Comme son nom l'indique.
**
**	Input : Le display, la fenetre de scroll
**	Ouput :
*/
void DeleteScroll ( display, swin )
    Display *display;
    SWin *swin;
{
    XDestroyWindow ( display, swin -> frame );
    (void) free ( (char *) swin );
}


/*
**	Function name : SetScrollLine
**
**	Description :  Positionne le nombre de ligne a scroller
**
**	Input :  Le display, nb lignes.
**	Ouput :
*/
void SetScrollLine ( s, n )
    SWin *s;
    int n;
{
    s -> line = n;
    s -> line_to_scroll = n - 1;

#ifdef DEBUG
    (void) fprintf ( stderr, "lines in buf = %d\n", s -> line );
#endif
}


/*
**	Function name : RefreshScroll
**
**	Description :  Ajuste la geometrie de la fenetre de scroll.
**		Positionne la scrollbar au bon endroit. Cette fonction
**		est utilisee apres un resize de la fenetre parent.
**
**	Input : Le display, le scroll, largeur, hauteur, no de la ligne courante.
**	Ouput :
*/
void RefreshScroll ( display, swin, width, height, n )
    Display *display;
    SWin	*swin;
    int 	width, height, n;
{
    swin -> f_height = height;
    swin -> fx = width - SCROLL_WIDTH;
    ShowScrollFrame ( display, swin );
    (void) MoveScrollBar ( display, swin, CURRENT, n );
}


/*
**	Function name : ShowScrollFrame
**
**	Description :  Affiche la fenetre de scroll en fonction de la
**		geometrie courante. Positionne la taille de la scrollbar.
**
**	Input :  Les display, le scroll
**	Ouput :
*/
void ShowScrollFrame ( display, swin )
    Display *display;
    SWin *swin;
{
  XMoveResizeWindow ( display, 
		       swin -> frame,
		       swin -> fx,
		       swin -> fy,
		       swin -> width,
		       swin -> f_height );
    
    XMoveWindow ( display, swin -> up, 2, 2 );
    XMoveWindow ( display, swin -> down,
		 2, swin -> f_height - SCROLL_BUTTONS_HEIGHT - 2 );
    
    XMapSubwindows ( display, swin -> frame ); 
    XMapWindow ( display, swin -> frame );
    SetScrollBarSize ( display, swin );
}


/*
**	Function name : SetScrollBarSize
**
**	Description :  Calcul de la taille de la scrollbar.
**
**	Input : Le display, le scroll.
**	Ouput :
*/
void SetScrollBarSize ( display, swin )
    Display *display;
    SWin *swin;
{
    double x;
    
    x = ((double) swin -> linepage) 
      / ((double) swin -> line_to_scroll + swin -> linepage);
    
    swin -> s_height =
      (int)( myrint ((double)(swin -> f_height - 4 - 2*SCROLL_BUTTONS_HEIGHT) * x));
    
    if ( swin -> s_height < DEFAULT_SIZE ) {
#ifdef DEBUG
	(void) fprintf ( stderr, "s_height 10\n" );
#endif
	swin -> s_height = DEFAULT_SIZE;
    }
    XResizeWindow ( display, swin -> scroll, SCROLL_WIDTH - 4, swin -> s_height ); 
    
    XMapWindow ( display, swin -> scroll );
    
    if ( swin -> line_to_scroll != 0 )
      swin -> delta = 
	((double)( swin -> f_height - 4 - 2*SCROLL_BUTTONS_HEIGHT - swin -> s_height ) 
	 / (double)(swin -> line_to_scroll));
    else
      swin -> delta = 0;	
#ifdef DEBUG
    (void) fprintf ( stderr, "delta = %.2f n = %.2f\n", 
		    swin -> delta,
		    ( swin -> f_height - 4 - 2*SCROLL_BUTTONS_HEIGHT - swin -> s_height )
		    / swin -> delta);
#endif
}


/*
**	Function name : ExposeInScroll
**
**	Description : Traitement d'un expose event.
**	Input : Le display, la fenetre exposee, le scroll courant.
**	Ouput : Vrai si le courant est expose faux sinon.
*/
int ExposeInScroll ( dpy, w, swin )
    Display *dpy;
    Window w;
    SWin *swin;
{
    XEvent ev;
    
    if ( w == swin -> frame ) {	
	RefreshScrollFrame ( dpy, swin );
	while ( XCheckWindowEvent ( dpy, w,ExposureMask, &ev ));
	return True;
    }
    if ( w == swin -> scroll ) {
	RefreshScrollBar ( dpy, swin );
	while ( XCheckWindowEvent ( dpy, w, ExposureMask, &ev ));
	return True;
    }
    return False;
}

/*
**	Function name : ButtonPressInScroll
**
**	Description :  Comme son nom l'indique.
**
**	Input : le scroll, le fenetre qui a recu ButtonPress, y position.
**	Ouput : True or False.
*/
int ButtonPressInScroll ( swin, w, y, result )
    SWin *swin;
    Window w;
    int y;
    int *result; /* RETURN */
{
    if ( swin -> frame != w )
      return False;
    
    if ( y < swin -> sy ) 
      *result = PREVIOUS;
    else if ( y > swin -> s_height + swin -> sy ) 
      *result = NEXT;
    else {
      swin -> y_max = swin -> f_height - 2 - SCROLL_BUTTONS_HEIGHT
	- ( swin -> sy + swin -> s_height - y );
      swin -> y_min = y - swin -> sy + 2 + SCROLL_BUTTONS_HEIGHT;
      swin -> last_y = y;
#ifdef DEBUG
      (void) fprintf ( stderr, "y = %d, sy = %d, fy = %d => y_max = %d y_min = %d\n",
		      y, swin->sy, swin->fy, swin -> y_max, swin -> y_min );
#endif
      *result = CURSOR;
    }
    return True;
}

/*
**	Function name : IsScrollButton 
**
**	Description :  Comme son nom l'indique.
**
**	Input : le scroll, le fenetre qui a recu ButtonPress, y position.
**	Ouput : 0 si up, 1 si down, sinon -1
*/
int IsScrollButton ( swin, w )
    SWin *swin;
    Window w;
{
    if ( swin -> up == w )
      return 0;
    else if ( swin -> down == w )
      return 1;
    else
      return -1;
}

void Display3DArrow ( s, w, run)
    SWin *s;
    Window w;
    int run;
{
  extern Display *dpy;
  
  if (w == s -> down) {
    if (run==1){
      XUnmapWindow (dpy,s -> down);
      XSetWindowBackgroundPixmap ( dpy, s -> down, sb_down_pix_run );
      XMapWindow (dpy,s -> down);
    }
    else {
      XUnmapWindow (dpy,s -> down);
      XSetWindowBackgroundPixmap ( dpy, s -> down, sb_down_pix );
      XMapWindow (dpy,s -> down);
    }
  }
  else {
    if (run==1){
      XUnmapWindow (dpy,s -> up);
      XSetWindowBackgroundPixmap ( dpy, s -> up, sb_up_pix_run );
      XMapWindow (dpy,s -> up);
    }
    else {
      XUnmapWindow (dpy,s -> up);
      XSetWindowBackgroundPixmap ( dpy, s -> up, sb_up_pix );
      XMapWindow (dpy,s -> up);
    }
  }
}

/*
**	Function name : HandleScrollBar
**
**	Description :  Synchronise la scrollbar avec le defilement du texte pendant
**		le ButtonPress.
**		Pas simple mes amis, pas simple. 
**
**	Input :  Le display, le scroll, la callback.
**	Ouput :
*/
void HandleScrollBar ( display, swin, f )
    Display *display;
    SWin *swin;
    void (*f) ();
{
    int n, dy;
    double dn;
    XEvent event;
    Window root, child;
    unsigned int keys_buttons;
    int root_x, root_y, win_x, win_y, stat = 0;
    
    while ( XCheckMaskEvent ( display, PointerMotionMask, &event ));
    XDefineCursor ( display, swin -> frame, cursor );
    Display3D ( display, swin -> scroll, swin -> bg , swin -> fg , 1, DOWN );
    
/*    XGrabPointer ( display, swin -> frame, True, ButtonReleaseMask | PointerMotionMask,
		  GrabModeAsync, GrabModeAsync, swin -> frame, None, CurrentTime );
*/    
#ifdef DEBUG
    fprintf ( stderr, "delta = %.3f f_height = %d s_height = %d\n",
	     swin -> delta, swin -> f_height, swin -> s_height );
#endif
    dn = swin -> rest;
    for ( ;; ) {
	XNextEvent ( display, &event );
	switch ( event.type ) {
	case ButtonRelease:
/*	  XUngrabPointer ( display, CurrentTime ); */
	  XUndefineCursor ( display, swin -> frame );
	  swin -> last_y = 0;
	  swin -> rest = dn;
#ifdef DEBUG
	  fprintf ( stderr, "dn = %.2f rest = %.2f\n", dn, swin -> rest );
#endif
	  return;
	case MotionNotify:
	  while ( XCheckMaskEvent ( display,
				   ButtonMotionMask, &event ));
	  XQueryPointer ( display,
			 event.xmotion.window, &root, &child, &root_x,
			 &root_y, &win_x, &win_y, &keys_buttons );
	  if ( (dy = win_y - swin -> last_y ) == 0 ) continue;
	  if ( dy > 0 ) {
	      if ( stat != BOTTOM ) {
		  if ( (stat = MoveScrollBar ( display,
					      swin, DOWN, win_y - swin -> last_y )) == BOTTOM ) {
		      win_y = swin -> y_max;
		      swin -> rest = 0;
		      dy = win_y - swin -> last_y;
		  }
		  dn += (dy / swin -> delta);
		  n = (int) (myrint ((double)dn));  
		  if ( n > 0 ) {
		      swin -> rest = (double) (dn - n);
		      ( * f ) ( n, swin -> text );
		      dn = swin -> rest;
		  }
		  swin -> last_dir = BOTTOM;
		  swin -> last_y = win_y;
	      }
	      else 
		continue;
	  }
	  else {
	      if ( stat != TOP ) {
		  if ( (stat = MoveScrollBar ( display,
					      swin, UP, swin -> last_y - win_y )) == TOP ) {
		      win_y = swin -> y_min;
		      swin -> rest = 0;
		      dy = win_y - swin -> last_y;
		  }
		  dn += dy / swin -> delta;
		  n = (int) (myrint ((double)dn));
		  if ( n < 0 ) {
		      swin -> rest = (double) (dn - n);
		      ( * f ) ( n, swin -> text );
		      dn = swin -> rest;
		  }
		  swin  -> last_dir = TOP;
		  swin -> last_y = win_y;
	      }
	      else
		continue;
	  }
#ifdef DEBUG
	  (void) fprintf ( stderr, "dy = %d\tlast = %d\n", dy, swin -> last_y );
#endif
	  
#ifdef DEBUG
	  (void) fprintf ( stderr, "dy = %d\tdn = %.2f\tn = %d\t rest = %.2f\n",
			  dy, dn , n, swin -> rest );
#endif
	  break;
	}
    }
    Display3D ( display, swin -> scroll, swin -> bg , swin -> fg , 1, UP );
}

/*
**	Function name : MoveScrollBar
**
**	Description :  Deplacement de la scrollbar suivant la valeur du
**		flag ( 1er page, page suivante... etc ).
**
**	Input :  Le display, le scroll, flags, deplacement en pixels
**	Ouput : L'etat courant.
*/
int MoveScrollBar ( display, swin, flag, dy )
    Display *display;
    SWin *swin;
    int flag, dy;
{
  int stat = 0;
  int i=0, j, n=0;
  double dn = swin -> rest;
  double d = 1.0 / swin -> delta;
  
#ifdef DEBUG
  fprintf ( stderr, "dy = %d\n", dy );
#endif
  if ( swin -> sy == 2 + SCROLL_BUTTONS_HEIGHT) {
    stat = TOP;
    swin -> rest = 0;
  }
  if ( swin -> sy == 
      swin -> f_height - 2 - SCROLL_BUTTONS_HEIGHT - swin -> s_height ) {
    stat = BOTTOM;
    swin -> rest = 0;
  }
  switch ( flag ) {
  case UP:
    if ( stat == TOP )
      return stat;
    if ( swin -> sy - dy < (2 + SCROLL_BUTTONS_HEIGHT) ) {
      swin -> sy = 2 + SCROLL_BUTTONS_HEIGHT;
      stat = TOP;
    }
    else {
      swin -> sy -= dy;
      stat = OTHER;
    }
    break;
  case DOWN:
    if ( stat == BOTTOM )
      return stat;
    if ( (swin -> sy + dy) > (swin -> f_height - 2 - SCROLL_BUTTONS_HEIGHT - swin -> s_height)) {
      swin -> sy = swin -> f_height - 2 - SCROLL_BUTTONS_HEIGHT - swin -> s_height;
      stat = BOTTOM;
    }
    else {
      swin -> sy += dy;
      stat = OTHER;
    }
    break;
  case FIRST:
    swin -> sy = 2 + SCROLL_BUTTONS_HEIGHT;
    swin -> rest = 0;
    stat = TOP;
    break;
  case NEXT:
    j = swin -> linepage - 1;
    while ( n < j ) {
      i++;
      dn += d;
      n = (int) (myrint ((double)dn));
    }
    swin -> sy += i;
    swin -> rest = (double) (dn - j);
#ifdef DEBUG
    (void) fprintf ( stderr, "i = %d rest = %.3f dn = %.3f n = %d \n",
		    i, swin -> rest, dn, n );
#endif
    if ( swin -> sy > (swin -> f_height - 2 - SCROLL_BUTTONS_HEIGHT - swin -> s_height) ) {
      swin -> sy = swin -> f_height - 2 - SCROLL_BUTTONS_HEIGHT - swin -> s_height;
      swin -> rest = 0;
    }
    stat = OTHER;
    break;
  case PREVIOUS:
    j = swin -> linepage - 1;
    while ( -n < j ) {
      i ++;
      dn -= d;
      n = (int) myrint ((double)dn);
    }
    swin -> sy -= i;
    swin -> rest = (double) (dn - (-j));
#ifdef DEBUG
    (void) fprintf ( stderr, "i = %d rest = %.3f dn = %.3f n = %d \n",
		    i, swin -> rest, dn, n );
#endif
    if ( swin -> sy < (2 + SCROLL_BUTTONS_HEIGHT) ) {
      swin -> sy = 2 + SCROLL_BUTTONS_HEIGHT;
      swin -> rest = 0;
    }
    stat = OTHER;
    break;
  case CURRENT:
    /* Le nombre de lignes qu'il aurait fallu scroller a partir du debut.  */
    j = dy;
    dn = 0;
#ifdef DEBUG
    (void) fprintf ( stderr,"j = %d\n", j );
#endif
    if ( dy == 0 )
      return 0;
    while ( n < j ) {
      i++;
      dn += d;
      n = (int) (myrint ((double)dn));
    }
    swin -> sy = i + 2 + SCROLL_BUTTONS_HEIGHT;
    swin -> rest = (double) (dn - j);
    break;
  }       
  XMoveWindow ( display, swin -> scroll, swin -> sx, swin -> sy );
  XFlush ( display );
  return stat;
}


/*
   **	Function name : RefreshScrollBar
   **
   **	Description :  Affichage du 3D pour la scrollbar.
   **
   **	Input : Le display, le scroll
   **	Ouput :
*/
void RefreshScrollBar ( display, s )
    Display *display;
    SWin *s;
{
  Display3D ( display, s -> scroll, s -> bg , s -> fg , 1, UP );
}

/*
   **	Function name : RefreshScrollFrame
   **
   **	Description :  Affichage du 3D pour la fenetre de scroll.
   **
   **	Input : Le display, le scroll
   **	Ouput :
*/
void RefreshScrollFrame ( display, s )
    Display *display;
    SWin *s;
{
  Display3D ( display, s -> frame, s -> bg , s -> fg , 2, DOWN );
}

/*
   **	Function name : myrint
   **
   **	Description :  Calcul de partie entiere. 
   **
   **	Input :  un reel signe
   **	Ouput : sa partie entiere
*/
static int myrint ( x )
    double x;
{
  int i = (int) x;
  int sign = (x >= 0) ? 1 : -1;
  double tmp = ( x > i ) ? ( x - i ) : ( i - x );
  
  if ( (0x1 & i ) != 0 ) {
    /* Impair */
    if ( tmp >= 0.5 ) i += sign;
  }
  else {
    /* Pair */
    if ( tmp > 0.5 ) i += sign;
  }
  return i;
}
